-------------------------------------------------------------------------------
-- Carrier sense CRS not used, because full duplex only
-- FULL DUPLEX only and NO flow control
-- no destination address filtering

-- Error flagged in the following conditions (from UG777)
--  FCS errors occur.

--  Packets are shorter than 64 bytes (undersize or fragment frames).

--  Jumbo frames are received when jumbo frames are not enabled.

--  VLAN frames of length 1519-1522 are received when VLAN frames are not enabled.

--  A frame above the programmed Max Frame Size is received when Max frame length
-- checking is enabled.

-- A value of 0x0000 to 0x002D is in the type/length field. In this situation, the frame
-- should be padded to minimum length. If it is not padded to exactly minimum frame
-- length, the frame is marked as bad (when length/type checking is enabled).

 -- A value of 0x002E to 0x0600 is in the type/length field, but the real length of the
-- received frame does not match this value (when length/type checking is enabled).

-- Any control frame that is received is not exactly the minimum frame length (unless
-- control frame length checks are disabled: see "Receiving a Pause Control Frame" on
-- page 82).

-- An error is indicated on the phy interface at any point during frame reception.

-- An error code is received in the 1-Gigabit frame extension field.

-- A valid pause frame, addressed to the MAC, is received when flow control is enabled.
-- See Overview of Flow Control.

--  A frame does not match against any of the enabled frame filters, if present.



-- O. Bourrion 12/2013
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.pack_mac.all;

entity mac_rx_engine is
port (
	rstn : in std_logic;
	clk : in std_logic;
	
	rx_tdata : out std_logic_vector(7 downto 0);
	rx_tvalid : out std_logic;
	rx_tlast : out std_logic;
	rx_tuser : out std_logic;
	rx_statistics_vector : out std_logic_vector(27 downto 0);
	rx_statistics_valid : out std_logic;
	rx_mac_config_vector : in std_logic_vector(79 downto 0);
	
	gmii_rxd : in std_logic_vector(7 downto 0);
	gmii_rx_dv : in std_logic;
	gmii_rx_er : in std_logic
	);
end mac_rx_engine;

architecture RTL of mac_rx_engine is
component crc_gen
generic (is_check : boolean :=false);
port(
	rstn : in std_logic;
	clk : in std_logic;
	clear_crc : in std_logic;

	data_en : in std_logic;
	data_in : in std_logic_vector(7 downto 0);

	crc_is_good : out std_logic;
	checksum: out std_logic_vector(7 downto 0)
);
end component;

type def_state_rx is (idle,preamble,SFD,data_xfer,remove_pad,wait_FCS,FCS,error_end);
signal state_rx : def_state_rx;

signal byte_cnt : natural range 0 to maxFrameSize-4;
signal payload_cnt : natural range 0 to maxFrameSize-18;
signal fcs_cnt : natural range 0 to 3;
signal length_received : std_logic;

signal length_field : std_logic_vector(15 downto 0);
signal length_field_type_1st : std_logic_vector(7 downto 0);
-- signal data_saved : std_logic_vector(7 downto 0);
signal checksum : std_logic_vector(7 downto 0);

signal crc_en : std_logic;
signal clear_crc : std_logic;

signal gmii_er_detected : std_logic;
signal FCS_error_detected : std_logic;

signal data_dlyd : std_logic_vector(7 downto 0);

constant dly_val : natural:=4;

type def_delay_tdata is array (0 to dly_val-1) of std_logic_vector(7 downto 0);
signal tdata_dlyd : def_delay_tdata;
signal valid_dlyd : std_logic_vector(dly_val-1 downto 0);

constant dly_gmii_val : natural:=4;
signal rx_dv_dlyd : std_logic_vector(dly_gmii_val-1 downto 0);
signal rx_er_dlyd : std_logic_vector(dly_gmii_val-1 downto 0);
signal prev_rx_tvalid : std_logic;
signal prev_rx_tdata : std_logic_vector(7 downto 0);
signal s_rx_tlast : std_logic;
signal crc_is_good : std_logic;

signal good_frame : std_logic;
signal bad_frame : std_logic;
signal broadcast_frame : std_logic;
signal ethertype : std_logic;
-- signal VLAN_frame : std_logic;

type def_mac_add is array (0 to 5) of std_logic_vector(7 downto 0);
signal s_DA : def_mac_add;
constant broadcast_const : def_mac_add:=(others=>(others=>'1'));
signal data_ethertype : std_logic_vector(7 downto 0);

begin
rx_tvalid<=(valid_dlyd(dly_val-1) and crc_en) or s_rx_tlast;
rx_tlast<=s_rx_tlast;
process(rstn,clk)
begin
	if rstn='0' then
		byte_cnt<=0;
		payload_cnt<=0;
		length_received<='0';
		fcs_cnt<=0;
		crc_en<='0';
		length_field<=(others=>'1');
		length_field_type_1st<=(others=>'0');
		prev_rx_tdata<=(others=>'0');
		rx_tdata<=(others=>'0');
		prev_rx_tvalid<='0';
-- 		rx_tvalid<='0';
		s_rx_tlast<='0';
		rx_tuser<='0';
		clear_crc<='1';
		gmii_er_detected<='0';
		data_dlyd<=(others=>'0');
		s_DA<=(others=>(others=>'0'));
		rx_dv_dlyd<=(others=>'0');
		rx_er_dlyd<=(others=>'0');
		FCS_error_detected<='0';
		rx_statistics_valid<='0';
		broadcast_frame<='0';
		ethertype<='0';
-- 		VLAN_frame<='0';
		valid_dlyd<=(others=>'0');
		tdata_dlyd<=(others=>(others=>'0'));
		data_ethertype<=(others=>'0');
		good_frame<='0';
		bad_frame<='0';
		state_rx<=idle;

	elsif rising_edge(clk) then
		data_dlyd<=gmii_rxd;
		if (state_rx/=data_xfer and ethertype='1') then
			rx_tdata<=data_ethertype;
		else
			rx_tdata<=tdata_dlyd(dly_val-1);
		end if;


		rx_dv_dlyd(0)<=gmii_rx_dv;
		rx_er_dlyd(0)<=gmii_rx_er;
		for i in 1 to dly_gmii_val-1 loop
			rx_dv_dlyd(i)<=rx_dv_dlyd(i-1);
			rx_er_dlyd(i)<=rx_er_dlyd(i-1);
		end loop;

		-- delay output
		valid_dlyd(0)<=prev_rx_tvalid;
		tdata_dlyd(0)<=prev_rx_tdata;
		for i in 1 to dly_val-1 loop
			valid_dlyd(i)<=valid_dlyd(i-1);
			tdata_dlyd(i)<=tdata_dlyd(i-1);
		end loop;

		
		if (s_DA=broadcast_const) then
			broadcast_frame<='1';
		else
			broadcast_frame<='0';
		end if;

		case state_rx is
			when idle =>
				FCS_error_detected<='0';
				byte_cnt<=0;
				payload_cnt<=0;
				length_field<=(others=>'1');
				length_field_type_1st<=(others=>'0');
				crc_en<='0';
				length_received<='0';
				clear_crc<='1';
				gmii_er_detected<='0';
				fcs_cnt<=0;
				s_rx_tlast<='0';
				rx_tuser<='0';
				prev_rx_tvalid<='0';
				rx_statistics_valid<='0';
				ethertype<='0';
-- 				VLAN_frame<='0';
				good_frame<='0';
				bad_frame<='0';

				if (gmii_rx_dv='1' and gmii_rxd=X"55") then
					state_rx<=preamble;
				end if;

			when preamble =>
				clear_crc<='0';
				if (gmii_rx_dv='0') then
					gmii_er_detected<='1';
					state_rx<=error_end;
				elsif (gmii_rx_er='1') then
					gmii_er_detected<='1';
					state_rx<=error_end;
				elsif (gmii_rxd=X"D5") then
					state_rx<=SFD;
				elsif (gmii_rxd=X"55") then
					state_rx<=preamble;
				end if;
			
			when SFD =>
				if (gmii_rx_dv='0') then
					gmii_er_detected<='1';
					state_rx<=error_end;
				elsif (gmii_rx_er='1') then
					gmii_er_detected<='1';
					state_rx<=error_end;
				else
					prev_rx_tdata<=gmii_rxd;
					s_DA(0)<=gmii_rxd;
					byte_cnt<=byte_cnt+1;
					crc_en<='1';
					state_rx<=data_xfer;
				end if;
			
			when data_xfer =>
				if (byte_cnt<6) then
					s_DA(byte_cnt)<=gmii_rxd;
				end if;
				if (byte_cnt=12) then --MSB length (IEEE802.3 see 3.2.6)
					length_field_type_1st(7 downto 0)<=gmii_rxd;
				end if;
				if (byte_cnt=13) then --LSB length
					length_field<=length_field_type_1st & gmii_rxd;
-- 					if (length_field_type_1st=X"81") then
-- 						VLAN_frame<='1';
-- 					end if;
					if (length_field_type_1st>=X"06") then
						ethertype<='1';
					else
						length_received<='1';
					end if;
				end if;

				if (gmii_rx_dv='0') then
					crc_en<='0';
					if (byte_cnt<13) then
						gmii_er_detected<='1';
						state_rx<=error_end;
						prev_rx_tvalid<='0';
						state_rx<=wait_FCS;
					elsif (byte_cnt<minFrameSize) then
						payload_cnt<=payload_cnt+1;
						state_rx<=error_end;
					else
						if (ethertype='1') then
							data_ethertype<=tdata_dlyd(dly_val-1);
							payload_cnt<=payload_cnt-4;
						end if;
						state_rx<=wait_FCS;
					end if;
				else
					if (gmii_rx_er='1') then -- ignore false carrier detection and normal frame reception with extension field (see p218 of pg047-gig-eth-pcs-pma)
						gmii_er_detected<='1';
						state_rx<=error_end;
					else
						byte_cnt<=byte_cnt+1;
						if (length_received='1') then
							if (payload_cnt<conv_integer(length_field)) then
								payload_cnt<=payload_cnt+1;
								prev_rx_tvalid<='1';
								prev_rx_tdata<=gmii_rxd;
							else
								if (byte_cnt<minFrameSize) then
									prev_rx_tvalid<='0';
									payload_cnt<=payload_cnt+1;
									state_rx<=remove_pad;
								end if;
							end if;
						end if;

						if (ethertype='1') then
							payload_cnt<=payload_cnt+1;
							prev_rx_tdata<=gmii_rxd;
							prev_rx_tvalid<='1';
						end if;
						if (ethertype='0' and length_received='0') then
							prev_rx_tdata<=gmii_rxd;
							prev_rx_tvalid<='1';
						end if;
					end if;
				end if;

			when remove_pad =>
				if (gmii_rx_dv='0') then
					crc_en<='0';
					payload_cnt<=payload_cnt-4;
					if (byte_cnt=minFrameSize) then
						state_rx<=wait_FCS;
					else
						state_rx<=error_end;
					end if;
				else
					payload_cnt<=payload_cnt+1;
					byte_cnt<=byte_cnt+1;
					if (gmii_rx_er='1') then
						gmii_er_detected<='1';
						state_rx<=error_end;
					end if;
				end if;


			when wait_FCS =>
				prev_rx_tvalid<='0';
				if (fcs_cnt=1) then
					fcs_cnt<=0;
					state_rx<=FCS;
				else
					fcs_cnt<=fcs_cnt+1;
				end if;

			when FCS =>
				if (rx_dv_dlyd(dly_gmii_val-1)='0') then
					gmii_er_detected<='1';
					state_rx<=error_end;
				elsif (rx_er_dlyd(dly_gmii_val-1)='1') then
					gmii_er_detected<='1';
					state_rx<=error_end;
				else
					if (crc_is_good='0') then
						FCS_error_detected<='1';
						state_rx<=error_end;
					else
						s_rx_tlast<='1';
						prev_rx_tvalid<='1';
						good_frame<='1';
						rx_statistics_valid<='1';
						state_rx<=idle;
					end if;
				end if;

			when error_end =>
				FCS_error_detected<='0';
				if (gmii_rx_dv='0') then -- no avoid continuation if it rx_err only
					s_rx_tlast<='1';
					prev_rx_tvalid<='1';
					if (fcs_error_detected='1' or gmii_er_detected='1') then
						bad_frame<='1';
						rx_tuser<='1';
					else
						good_frame<='1';
					end if;
					rx_statistics_valid<='1';
					state_rx<=idle;
				end if;
		end case;
	end if;
end process;

rx_statistics_vector(0)<=good_frame;
rx_statistics_vector(1)<=bad_frame;
rx_statistics_vector(2)<=fcs_error_detected;
rx_statistics_vector(3)<=broadcast_frame;
rx_statistics_vector(4)<=s_DA(0)(0); -- multicast frame bit 0 first field at 0 sect3.2.3 IEEE802.3
rx_statistics_vector(18 downto 5)<=conv_std_logic_vector(byte_cnt,14);
rx_statistics_vector(24 downto 19)<=(others=>'0');
rx_statistics_vector(25)<='1' when (conv_integer(length_field)/=payload_cnt or payload_cnt<46) else '0'; -- Length out of range
rx_statistics_vector(26)<='0'; -- ALIGNEMENT ERROR (for speed < 1Gb/s)
rx_statistics_vector(27)<='0'; -- ADDRESS_MATCH

lbl_crc : crc_gen generic map(is_check=>true)
	port map(
	rstn=>rstn,
	clk=>clk,
	clear_crc=>clear_crc,

	data_en=>crc_en,
	data_in=>data_dlyd,

	crc_is_good=>crc_is_good,
	checksum=>checksum
);

end RTL;

