library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

entity mac_fifo_axi4_lpsc is
  port (
    m_aclk : in std_logic;
    s_aclk : in std_logic;
    s_aresetn : in std_logic;
    s_axis_tvalid : in std_logic;
    s_axis_tready : out std_logic;
    s_axis_tdata : in std_logic_vector(7 downto 0);
    s_axis_tlast : in std_logic;
    s_axis_tuser : in std_logic_vector(0 downto 0);
    m_axis_tvalid : out std_logic;
    m_axis_tready : in std_logic;
    m_axis_tdata : out std_logic_vector(7 downto 0);
    m_axis_tlast : out std_logic;
    m_axis_tuser : out std_logic_vector(0 downto 0)
  );
end mac_fifo_axi4_lpsc;

architecture RTL of mac_fifo_axi4_lpsc is
constant add_size : natural :=10; -- mem_depth is power of 2
type def_mem is array (0 to 2**add_size-1) of std_logic_vector (9 downto 0);
signal mem : def_mem;

type def_state_fwft is (is_empty, is_not_empty);
signal state_fwft : def_state_fwft;

signal read_enable:           std_logic;
signal read_enable_1ck:       std_logic;
signal dsel:			       std_logic;
signal full:                  std_logic;
signal empty:                 std_logic;
signal read_addr:             std_logic_vector(add_size-1 downto 0);
signal read_addrgray:         std_logic_vector(add_size-1 downto 0);
signal read_nextgray:         std_logic_vector(add_size-1 downto 0);
signal read_lastgray:         std_logic_vector(add_size-1 downto 0);
signal write_addr:            std_logic_vector(add_size-1 downto 0);
signal write_addrgray:        std_logic_vector(add_size-1 downto 0);
signal write_nextgray:        std_logic_vector(add_size-1 downto 0);
signal emptyg:                std_logic;
signal fullg:                 std_logic;
signal memout:		          std_logic_vector(9 downto 0);
signal memout_1ck:	          std_logic_vector(9 downto 0);
signal memout_fwft:	          std_logic_vector(9 downto 0);

function conv_seq_to_gray(input : std_logic_vector) return std_logic_vector is
variable int_vect : std_logic_vector(input'length-1 downto 0);
begin
	for i in 0 to input'length-2 loop
		int_vect(i):=input(i+1) XOR input(i);
	end loop;
	int_vect(input'length-1):=input(input'length-1);

return int_vect;		
end conv_seq_to_gray;


function conv_gray_to_seq(input : std_logic_vector) return std_logic_vector is
variable int_vect : std_logic_vector(input'length-1 downto 0);
variable int_xor : std_logic;
begin
	int_xor:=input(input'length-1);
	int_vect(input'length-1):=int_xor;
	
	for i in input'length-2 downto 0 loop
		int_xor:=int_xor xor input(i);
		int_vect(i):=int_xor;
	end loop;

return int_vect;
end conv_gray_to_seq;

constant MAX_min_one   : std_logic_vector(add_size-1 downto 0):=(others=>'1');
constant MAX_min_two   : std_logic_vector(add_size-1 downto 0):=(0=>'0',others=>'1');
constant MAX_min_three : std_logic_vector(add_size-1 downto 0):=(1=>'0',others=>'1');

BEGIN
--------------------------------------------------------------------------
-- write side (slave side)
--------------------------------------------------------------------------
process(s_aclk)
begin
	if rising_edge(s_aclk) then
		if s_axis_tvalid='1' then
			mem(conv_integer(write_addr))<=s_axis_tuser(0) & s_axis_tlast & s_axis_tdata;
		end if;
	end if;
end process;


write_proc: process (s_aclk, s_aresetn)
begin
if (s_aresetn = '0') then
	write_addr <= (others=>'0');
	write_nextgray <= conv_seq_to_gray(max_min_one);
	write_addrgray <=conv_seq_to_gray(max_min_two);
	full <= '0';

elsif rising_edge(s_aclk) then
	if (s_axis_tvalid='1' and full='0') then
		write_addr <= write_addr + 1;
		write_nextgray<=conv_seq_to_gray(write_addr);
		write_addrgray <= write_nextgray;
	end if;
	
	if (full='1' or s_axis_tvalid='1') then
		full <= fullg;
	end if;
	
end if;
end process write_proc;

s_axis_tready<=not full;

--------------------------------------------------------------------------
-- read side (master side)
--------------------------------------------------------------------------
process(m_aclk)
begin
	if rising_edge(m_aclk) then
		memout <=mem(conv_integer(read_addr));
	end if;
end process;

read_proc: process (m_aclk, s_aresetn)
begin
	if (s_aresetn = '0') then
		read_addr <= (others=>'0');
		read_nextgray <=conv_seq_to_gray(max_min_one);
		read_addrgray <=conv_seq_to_gray(max_min_two);
		read_lastgray <=conv_seq_to_gray(max_min_three);
		empty <= '1';

	elsif rising_edge(m_aclk) then
		if (read_enable='1' and empty='0') then
			read_addr <= read_addr + 1;
			read_nextgray <=conv_seq_to_gray(read_addr);
			read_addrgray <= read_nextgray;
			read_lastgray <= read_addrgray;
		end if;
		if (empty = '1' or read_enable='1') then
			empty <= emptyg;
		end if;

	end if;
end process read_proc;

emptyg<='1' when (write_addrgray=read_addrgray and empty='1') or (write_addrgray=read_nextgray and empty='0') else '0';
fullg<='1' when  (read_lastgray=write_addrgray and full='1')  or (read_lastgray=write_nextgray and full='0')  else '0';

--------------------------------------------------------------------------
-- read side (master side)
--------------------------------------------------------------------------
fwft : process(s_aresetn,m_aclk)
begin
	if s_aresetn='0' then
		memout_1ck<=(others=>'0');
		read_enable<='0';
		read_enable_1ck<='0';
		m_axis_tvalid<='0';
		dsel<='0';
		state_fwft<=is_empty;


	elsif rising_edge(m_aclk) then
		case state_fwft is
			when is_empty=>
				if empty='0' then
					read_enable<='1';
					state_fwft<=is_not_empty;
				else
					m_axis_tvalid<='0';
				end if;

			when is_not_empty =>
				if empty='1'or (write_addrgray=read_nextgray and empty='0') then
					m_axis_tvalid<='1';
					read_enable<='0';
					state_fwft<=is_empty;
				else
					read_enable<=m_axis_tready;
					m_axis_tvalid<='1';
				end if;
		end case;


		read_enable_1ck<=read_enable;
		dsel <= read_enable and read_enable_1ck;
-- 		dsel <= read_enable xnor read_enable_1ck;
			
		if read_enable = '1' or read_enable_1ck = '1' then
			memout_1ck<=memout;
		end if;
		
	end if;
end process fwft;

memout_fwft <= memout_1ck when dsel = '0' else memout;

m_axis_tuser(0)<=memout_fwft(9);
m_axis_tlast <=memout_fwft(8);
m_axis_tdata <=memout_fwft(7 downto 0);

END RTL;

