-- ntp_drift
--
-- computes local DDS adjustement to align frequency to reference
-- freq_adj=theta/(n*2^poll_interval)
-- 
-- This block assume that the large adjustements were already 
-- made during the initialisation sequence
-- 
-- => strong assumption made : offset is below the second 
-- => Other assumption freq_adj is in Hz (clock_precision do not play a role here)
-- Adjustement precision is hence at the Hz level

-- Olivier Bourrion, May 2015

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity ntp_drift is 
generic(
	poll_interval_exp :  integer:=7; -- 7 for 2^7=128s -- should be between 6 and 10
	clock_precision :  integer:=20 -- defined as 1s/2^-clock precision (20 for ~ 1 us)
	);
port(
	refclk: in std_logic;
	rst: in std_logic; -- rst
	forbid: in std_logic; -- rst
    
	update_local_time: in std_logic;
	theta : in std_logic_vector(63 downto 0);
	n : in std_logic_vector(5 downto 0); -- number of poll intervals since last theta update (between 0 and 7)x8

	update_dds : out std_logic;
	freq_adj : out std_logic_vector(clock_precision-1 downto 0)
	);
end ntp_drift;

architecture RTL of ntp_drift is
component serial_unsigned_division is
generic(
	dividend_size : natural :=32;
	divisor_size : natural :=4
	);
port(
	clk : in  std_logic;
	reset : in  std_logic;
	load  : in  std_logic;
	dividend : in  unsigned(dividend_size-1 downto 0);
	divisor : in  unsigned(divisor_size-1 downto 0);

	ready : out std_logic;
	quotient : out unsigned(dividend_size-1 downto 0);
	remain : out unsigned(dividend_size-1 downto 0)
	);
end component serial_unsigned_division;

constant sub_sec_bit : natural :=32-clock_precision;

signal upd_dly : std_logic;
signal start_division : std_logic;

signal theta_bin_div : unsigned(31 downto 0);
signal saved_n : unsigned(n'length downto 0);
signal theta_sign : std_logic;

signal dividend : unsigned(31 downto 0);
signal quotient : unsigned(31 downto 0);
signal freq_adj_s : unsigned(31 downto 0);

signal ready : std_logic;
signal new_freq_adj : std_logic_vector(clock_precision-1 downto 0);
signal update_average : std_logic;
-- constant average_size : natural :=4;
-- signal average_cnt : natural range 0 to average_size-1;
-- type def_array_32 is array (0 to average_size-1) of unsigned (31 downto 0);
-- signal freq_adj_array : def_array_32;
-- signal freq_adj_acc : unsigned(31 downto 0);
-- 
-- type def_state_average is (idle,compute,apply);
-- signal state_average : def_state_average;

begin

process(refclk)
begin
	if rising_edge(refclk) then
		-- Operations sequencing
		if rst='1' then
			upd_dly<='0';
			update_average<='0';
			start_division<='0';
		else
			upd_dly<=update_local_time;
			start_division<=upd_dly;
		end if;

		-- First binary division (right shift)
		if update_local_time='1' then
			-- divide by poll interval
-- 			theta_bin_div<=unsigned(theta(31+poll_interval_exp downto poll_interval_exp)); -- sign is kept if and only if theta is lower than poll_interval!!
-- 			theta_sign<=theta(31+poll_interval_exp);
			theta_bin_div<=unsigned(theta(31+sub_sec_bit downto sub_sec_bit)); -- sign is kept if and only if theta is lower than poll_interval!!
			theta_sign<=theta(31+sub_sec_bit);
			saved_n<=unsigned('0' & n)+1;
		end if;

		-- regular division preparation (for neg number)
		if upd_dly='1' then
			if theta_sign='1' then --neg number
				dividend<=not(theta_bin_div)+1;
			else
				dividend<=theta_bin_div;
			end if;
		end if;

		-- division result management
		if ready='1' then
			if  forbid='0' then
				if theta_sign='1' then --neg number
					freq_adj_s<=not(quotient)+1;
				else
					freq_adj_s<=quotient;
				end if;
			else
				freq_adj_s<=(others=>'0');
			end if;
		end if;
		update_average<=ready;
	end if;
end process;
-- freq_adj<=std_logic_vector(freq_adj_s(clock_precision-1 downto 0));
update_dds<=update_average;
freq_adj<=std_logic_vector(freq_adj_s(clock_precision+poll_interval_exp-1 downto poll_interval_exp));

-------------------------------------------------------------------------------
-- averager (solution 1)
-------------------------------------------------------------------------------
-- process(refclk)
-- begin
-- 	if rising_edge(refclk) then
-- 		-- Operations sequencing
-- 		if in_init1='1' then
-- 			update_dds<=update_average;
-- 			freq_adj<=std_logic_vector(freq_adj_s(clock_precision+poll_interval_exp-1 downto poll_interval_exp));
-- 			freq_adj_array<=(others=>(others=>'0'));
-- 			state_average<=idle;
-- 		else
-- 			case state_average is
-- 				when idle =>
-- 					freq_adj_acc<=(others=>'0');
-- 					update_dds<='0';
-- 					average_cnt<=0;
-- 					if update_average='1' then
-- 						freq_adj_array(0)<=freq_adj_s;
-- 						for i in 1 to average_size-1 loop
-- 							freq_adj_array(i)<=freq_adj_array(i-1);
-- 						end loop;
-- 						state_average<=compute;
-- 					end if;
-- 
-- 				when compute =>
-- 					freq_adj_acc<=freq_adj_acc+freq_adj_array(average_cnt);
-- 					if average_cnt=average_size-1 then
-- 						state_average<=apply;
-- 					else
-- 						average_cnt<=average_cnt+1;
-- 					end if;
-- 
-- 				when apply=>
-- 					freq_adj<=std_logic_vector(freq_adj_acc(clock_precision+poll_interval_exp-1+2 downto poll_interval_exp+2)); --is divided by 4
-- 					update_dds<='1';
-- 					state_average<=idle;
-- 			end case;
-- 		end if;
-- 	end if;
-- end process;
-------------------------------------------------------------------------------
-- allow freq_adj of +/-1 only (solution 2)
-------------------------------------------------------------------------------
-- process(refclk)
-- begin
-- 	if rising_edge(refclk) then
-- 		-- Operations sequencing
-- 		if in_init1='1' then
-- 			update_dds<=update_average;
-- 			freq_adj<=std_logic_vector(freq_adj_s(clock_precision+poll_interval_exp-1 downto poll_interval_exp));
-- 			state_average<=idle;
-- 		else
-- 			case state_average is
-- 				when idle =>
-- 					update_dds<='0';
-- 					if update_average='1' then
-- 						state_average<=compute;
-- 					end if;
-- 
-- 				when compute =>
-- 					if freq_adj_s(clock_precision+poll_interval_exp-1 downto poll_interval_exp)=0 then
-- 						freq_adj<=(others=>'0');
-- 					elsif freq_adj_s(clock_precision+poll_interval_exp-1)='1' then --neg number
-- 						freq_adj<=(others=>'1'); -- -1
-- 					else -- pos number
-- 						freq_adj<=(0=>'1',others=>'0'); -- +1
-- 					end if;
-- 					state_average<=apply;
-- 
-- 				when apply=>
-- 					update_dds<='1';
-- 					state_average<=idle;
-- 			end case;
-- 		end if;
-- 	end if;
-- end process;
-- 
-- 
-------------------------------------------------------------------------------
-- divider
-------------------------------------------------------------------------------
divider : serial_unsigned_division generic map(
	dividend_size=>32,
	divisor_size=>saved_n'length
	)
port map(
	clk=>refclk,
	reset=>rst,
	load=>start_division,
	dividend=>dividend,
	divisor=>saved_n,

	ready=>ready,
	quotient=>quotient,
	remain=>open
	);

end RTL;
