-- ntp_main
--
-- NTP client top level
--
-- clock precision is conditionned by :
--		2**clock_precision must be less than half of REFCLK_FREQ (otherwise DDS do not work)
--		REFCLK/8 must be larger than 2**clock_precision, otherwise ntp_local_time doesn't work optimally, clock ticks are missed on time_incr.
--		REFCLK/4 must be larger than 2**clock_precision, otherwise ntp_local_time does work at all.
-- Olivier Bourrion, May 2015

library ieee;
use ieee.std_logic_1164.all;
use IEEE.NUMERIC_STD.ALL;

use work.pack_ipv4_types.all;
use work.pack_axi.all;
use work.pack_arp_types.all;

entity ntp_main is 
generic(
	IN_SIMULATION :  boolean:=false; -- when true, time_enable period is  accelerated by 30000 (3 us instead of 1s)
	REFCLK_FREQ :  natural:=10E6; -- default to 10MHz
	en_freq_adj : boolean :=true; --! Enable freq adjustement (to put false if clocked with atomic clock)
	stratum :  natural:=7; -- Must be >1, 6 or above recommanded
	poll_interval_exp :  integer:=7; -- 7 for 2^7=128s -- should be between 6 and 10
	clock_precision :  integer:=20 -- defined as 1s/2^-clock precision (20 for ~ 1 us)
	);
port(
	mac_clk: in std_logic; -- Ethernet MAC clock (125MHz)
	rst_macclk: in std_logic; -- MAC clock domain sync reset

	our_ip_address 	: in std_logic_vector (31 downto 0);
	ntp_server_add 	: in std_logic_vector (31 downto 0);
	enable_ntp : in std_logic;

	-- UDP RX signals
    udp_rx_start : in std_logic;       -- indicates receipt of udp header
    udp_rxo      : in udp_rx_type;

	-- UDP TX signals
	udp_tx_start			: out std_logic;					-- indicates req to tx UDP
	udp_txi					: out udp_tx_type;					-- UDP tx cxns
	udp_tx_result			: in std_logic_vector (1 downto 0); -- tx status (changes during transmission)
	udp_tx_data_out_ready   : in std_logic;						-- indicates udp_tx is ready to take data

	-- UDP TX signals (from next stage)
	udp_tx_start_prior	    : in std_logic;
	udp_txi_prior			: in udp_tx_type;
	udp_tx_result_prior		: out std_logic_vector (1 downto 0);
	udp_tx_data_out_ready_prior: out std_logic;

	-- refclk interface (xtime slower than mac_clk) at least half (60MHz)
	refclk: in std_logic;
	rst_refclk: in std_logic;

	-- time and monitoring (refclk sync)
	new_time_avail: out std_logic;-- several refclk pulse on each update
	local_time : out std_logic_vector(63 downto 0);

	new_theta_delta: out std_logic; -- several refclk pulse on each update
	theta_i : out std_logic_vector(63 downto 0);
	delta_i : out std_logic_vector(63 downto 0);

	update_theta: out std_logic; -- several refclk pulse on each update
	theta : out std_logic_vector(63 downto 0);
	delta : out std_logic_vector(63 downto 0);
	freq_adj : out std_logic_vector(clock_precision-1 downto 0)
	);
end ntp_main;

architecture RTL of ntp_main is
component ntp_dds
generic(
	log_P: natural := 20; -- P is computed as 2**log_P
	Q: natural :=10E6
	);
port(
	refclk: in std_logic; -- Stable reference clock
	rst: in std_logic; -- rst
	
	upd_dds : in std_logic;
	freq_adj : in std_logic_vector(log_P-1 downto 0);
    
	sub_sec_top: out std_logic
	);
end component;

component ntp_clock_filter
port(
	refclk: in std_logic; -- Stable reference clock
	rst: in std_logic; -- rst
    
	update_tuple: in std_logic;
	is_valid: in std_logic;
	theta : in std_logic_vector(63 downto 0);
	delta : in std_logic_vector(63 downto 0);
	epoch  : in std_logic_vector(63 downto 0);

	update_local_time: out std_logic;
	allow_freq_adj: out std_logic;
	new_delta : out std_logic_vector(63 downto 0);
	new_theta : out std_logic_vector(63 downto 0);
	n : out std_logic_vector(5 downto 0);
	reference_time : out std_logic_vector(63 downto 0) -- last time when system was updated
	);
end component;
component ntp_theta_delta
port(
	refclk: in std_logic; -- Stable reference clock
	rst: in std_logic; -- rst
    
	ntp_frame_rcvd: in std_logic;
	ntp_timeout: in std_logic; --a time out if 4 successive poll failed
	T_1 : in std_logic_vector(63 downto 0); -- rcvd OT field
	T_2 : in std_logic_vector(63 downto 0); -- local time at datagram rcv time
	T_prime_1 : in std_logic_vector(63 downto 0); -- rcvd RT field
	T_prime_2 : in std_logic_vector(63 downto 0); -- rcvd TT field

	compute_done: out std_logic;
	is_valid: out std_logic;
	theta : out std_logic_vector(63 downto 0);
	delta : out std_logic_vector(63 downto 0);
	epoch  : out std_logic_vector(63 downto 0)
	);
end component ntp_theta_delta;

component ntp_drift
generic(
	poll_interval_exp :  integer:=7; -- 7 for 2^7=128s -- should be between 6 and 10
	clock_precision :  integer:=20 -- defined as 1s/2^-clock precision (20 for ~ 1 us)
	);
port(
	refclk: in std_logic;
	rst: in std_logic; -- rst
	forbid: in std_logic; -- rst
    
	update_local_time: in std_logic;
	theta : in std_logic_vector(63 downto 0);
	n : in std_logic_vector(5 downto 0); -- number of poll intervals since last theta update (between 0 and 7)

	update_dds : out std_logic;
	freq_adj : out std_logic_vector(clock_precision-1 downto 0)
	);
end component;

component ntp_local_time
generic(
   increment_bit: natural := 20 -- conputed form DDS log_P ->increment_bit=32-log_P
	);
port(
	refclk: in std_logic; -- Stable reference clock
	rst: in std_logic; -- rst
    
	sub_sec_top: in std_logic;
	update_theta: in std_logic;
	theta : in std_logic_vector(63 downto 0);

	new_time_avail: out std_logic; -- signal to use for clock domain synchronisation
	local_time : out std_logic_vector(63 downto 0)
	);
end component ntp_local_time;

component ntp_udp_client 
generic(
	IN_SIMULATION :  boolean:=false; -- when true, time_enable period is  accelerated by 30000 (3 us instead of 1s)
	REFCLK_FREQ :  natural:=10E6; -- default to 10MHz
	stratum :  natural:=7; -- Must be >1, 6 or above recommanded
	poll_interval_exp :  integer:=7; -- 7 for 2^7=128s -- should be between 6 and 10
	clock_precision :  integer:=20 -- defined as 1s/2^-clock precision (20 for ~ 1 us)
	);
port(
	mac_clk: in std_logic; -- Ethernet MAC clock (125MHz)
	rst_macclk: in std_logic; -- MAC clock domain sync reset

	our_ip_address 	: in std_logic_vector (31 downto 0);
	ntp_server_add 	: in std_logic_vector (31 downto 0);
	enable_ntp : in std_logic;

	-- UDP RX signals
    udp_rx_start : in std_logic;       -- indicates receipt of udp header
    udp_rxo      : in udp_rx_type;

	-- UDP TX signals
	udp_tx_start			: out std_logic;					-- indicates req to tx UDP
	udp_txi					: out udp_tx_type;					-- UDP tx cxns
	udp_tx_result			: in std_logic_vector (1 downto 0); -- tx status (changes during transmission)
	udp_tx_data_out_ready   : in std_logic;						-- indicates udp_tx is ready to take data

	-- UDP TX signals (from next stage)
	udp_tx_start_prior	    : in std_logic;
	udp_txi_prior			: in udp_tx_type;
	udp_tx_result_prior		: out std_logic_vector (1 downto 0);
	udp_tx_data_out_ready_prior: out std_logic;

	-- refclk interface (xtime slower than mac_clk)
	refclk: in std_logic;
	rst_refclk: in std_logic; -- referenced to refclk
	new_time_avail: in std_logic; -- signal to use for clock domain synchronisation
	local_time : in std_logic_vector(63 downto 0);
	reference_time : in std_logic_vector(63 downto 0); -- last time when system was updated

	ntp_frame_rcvd: out std_logic;
	ntp_timeout: out std_logic; --a time out if 4 successive poll failed
	T_1 : out std_logic_vector(63 downto 0); -- rcvd OT field
	T_2 : out std_logic_vector(63 downto 0); -- local time at datagram rcv time
	T_prime_1 : out std_logic_vector(63 downto 0); -- rcvd RT field
	T_prime_2 : out std_logic_vector(63 downto 0) -- rcvd TT field
	);
end component ntp_udp_client;

component ntp_pulse_stretcher
generic(
	stretch_val : natural
	);
port(
	clk : in std_logic;
	rst : in std_logic;
	
	pulse_in : in std_logic;
	pulse_out : out std_logic
	);
end component;
	
signal new_time_avail_s: std_logic;
signal local_time_s : std_logic_vector(63 downto 0);

signal ntp_frame_rcvd: std_logic;
signal ntp_timeout: std_logic;
signal T_1 : std_logic_vector(63 downto 0);
signal T_2 : std_logic_vector(63 downto 0);
signal T_prime_1 : std_logic_vector(63 downto 0);
signal T_prime_2 : std_logic_vector(63 downto 0);
signal sub_sec_top: std_logic;

signal update_theta_s: std_logic;
signal compute_done: std_logic;
signal is_valid: std_logic;
signal epoch  : std_logic_vector(63 downto 0);
signal new_theta : std_logic_vector(63 downto 0);
signal n : std_logic_vector(5 downto 0);
signal reference_time : std_logic_vector(63 downto 0);
signal new_reference_time : std_logic_vector(63 downto 0);
signal new_delta : std_logic_vector(63 downto 0);
signal theta_s : std_logic_vector(63 downto 0);
signal delta_s : std_logic_vector(63 downto 0);

signal update_dds : std_logic;
signal freq_adj_s : std_logic_vector(clock_precision-1 downto 0);
signal allow_freq_adj: std_logic;
signal forbid_freq_adj: std_logic;

begin
-------------------------------------------------------------------------------
-- NTP UDP INTERFACE
-------------------------------------------------------------------------------
lbl_ntp_udp : ntp_udp_client generic map(
	IN_SIMULATION=>IN_SIMULATION,
	REFCLK_FREQ=>REFCLK_FREQ,
	poll_interval_exp=>poll_interval_exp,
	clock_precision=>clock_precision
	)
port map(
	mac_clk=>mac_clk,
	rst_macclk=>rst_macclk,

	our_ip_address=>our_ip_address,
	ntp_server_add=>ntp_server_add,
	enable_ntp=>enable_ntp,

	-- UDP RX signals
    udp_rx_start=>udp_rx_start,
    udp_rxo=>udp_rxo,

	-- UDP TX signals
	udp_tx_start=>udp_tx_start,
	udp_txi=>udp_txi,
	udp_tx_result=>udp_tx_result,
	udp_tx_data_out_ready=>udp_tx_data_out_ready,

	-- UDP TX signals (from next stage)
	udp_tx_start_prior=>udp_tx_start_prior,
	udp_txi_prior=>udp_txi_prior,
	udp_tx_result_prior=>udp_tx_result_prior,
	udp_tx_data_out_ready_prior=>udp_tx_data_out_ready_prior,

	-- refclk interface (xtime slower than mac_clk)
	refclk=>refclk,
	rst_refclk=>rst_refclk,
	new_time_avail=>new_time_avail_s,
	local_time=>local_time_s,
	reference_time=>reference_time,

	ntp_frame_rcvd=>ntp_frame_rcvd,
	ntp_timeout=>ntp_timeout,
	T_1=>T_1,
	T_2=>T_2,
	T_prime_1=>T_prime_1,
	T_prime_2=>T_prime_2
	);
-------------------------------------------------------------------------------
-- NTP DDS
-------------------------------------------------------------------------------
lbl_ntp_dds : ntp_dds generic map(
	log_P=>clock_precision,
	Q=>REFCLK_FREQ
-- 	log_P=>20,
-- 	Q=>2**22
	)
port map(
	refclk=>refclk,
	rst=>rst_refclk,
    
	upd_dds=>update_dds,
	freq_adj=>freq_adj_s,
    
	sub_sec_top=>sub_sec_top
	);
-------------------------------------------------------------------------------
-- NTP LOCAL TIME
-------------------------------------------------------------------------------
lbl_ntp_local_time :  ntp_local_time generic map(increment_bit=>32-clock_precision)
port map(
	refclk=>refclk,
	rst=>rst_refclk,
    
	sub_sec_top=>sub_sec_top,
	update_theta=>update_theta_s,
	theta=>new_theta,

	new_time_avail=>new_time_avail_s,
	local_time=>local_time_s
	);

lbl_stretch_new_time_avail: ntp_pulse_stretcher generic map(
		stretch_val=>125E6/REFCLK_FREQ+1
		)
port map(
	clk=>refclk,
	rst=>rst_refclk,
	
	pulse_in=>new_time_avail_s,
	pulse_out=>new_time_avail
	);
local_time<=local_time_s;
-------------------------------------------------------------------------------
-- NTP THETA DELTA
-------------------------------------------------------------------------------
lbl_ntp_theta_delta : ntp_theta_delta port map(
	refclk=>refclk,
	rst=>rst_refclk,
    
	ntp_frame_rcvd=>ntp_frame_rcvd,
	ntp_timeout=>ntp_timeout,
	T_1=>T_1,
	T_2=>T_2,
	T_prime_1=>T_prime_1,
	T_prime_2=>T_prime_2,

	compute_done=>compute_done,
	is_valid=>is_valid,
	theta=>theta_s,
	delta=>delta_s,
	epoch=>epoch
	);
-- reference_time<=epoch;
-- theta<=theta_s;
-- delta<=delta_s;
-- update_theta<=compute_done;

theta_i<=theta_s;
delta_i<=delta_s;

lbl_stretch_new_theta_delta: ntp_pulse_stretcher generic map(
		stretch_val=>125E6/REFCLK_FREQ+1
		)
port map(
	clk=>refclk,
	rst=>rst_refclk,
	
	pulse_in=>compute_done,
	pulse_out=>new_theta_delta
	);
-------------------------------------------------------------------------------
-- NTP CLOCK FILTER
-------------------------------------------------------------------------------
lbl_clock_filter : ntp_clock_filter port map(
	refclk=>refclk,
	rst=>rst_refclk,
    
	update_tuple=>compute_done,
	is_valid=>is_valid,
	theta=>theta_s,
	delta=>delta_s,
	epoch=>epoch,

	update_local_time=>update_theta_s,
	allow_freq_adj=>allow_freq_adj,
	new_delta=>new_delta,
	new_theta=>new_theta,
	n=>n,
	reference_time=>new_reference_time
	);
forbid_freq_adj<=not(allow_freq_adj);

reference_time<=new_reference_time;
theta<=new_theta;
delta<=new_delta;

lbl_stretch_upd_theta: ntp_pulse_stretcher generic map(
		stretch_val=>125E6/REFCLK_FREQ+1
		)
port map(
	clk=>refclk,
	rst=>rst_refclk,
	
-- 	pulse_in=>update_theta_s,
	pulse_in=>update_dds, -- to save freq_adj corresponding to the theta at the same sample
	pulse_out=>update_theta
	);
-------------------------------------------------------------------------------
-- NTP DRIFT
-------------------------------------------------------------------------------
lbl_gen_freq_adj : if en_freq_adj generate
	lbl_ntp_drift : ntp_drift generic map(
		poll_interval_exp=>poll_interval_exp,
		clock_precision=>clock_precision
		)
	port map(
		refclk=>refclk,
		rst=>rst_refclk,
		forbid=>forbid_freq_adj,
		
		update_local_time=>update_theta_s,
		theta=>new_theta,
		n=>n,

		update_dds=>update_dds,
		freq_adj=>freq_adj_s
		);
	freq_adj<=freq_adj_s;
end generate lbl_gen_freq_adj;

lbl_rm_freq_adj : if not(en_freq_adj) generate
	freq_adj<=(others=>'0');
	update_dds<=update_theta_s;
end generate lbl_rm_freq_adj;


end RTL;
