--------------------------------------------------
-- ntp_theta_delta
-- 
-- Theta deta computer
-- epsilon not saved, not designed to work with several servers
-- FPGA ressources are limited
-- 
-- O. Bourrion May 2015
--------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity ntp_theta_delta is 
port(
	refclk: in std_logic; -- Stable reference clock
	rst: in std_logic; -- rst
    
	ntp_frame_rcvd: in std_logic;
	ntp_timeout: in std_logic; --a time out if 4 successive poll failed
	T_1 : in std_logic_vector(63 downto 0); -- rcvd OT field
	T_2 : in std_logic_vector(63 downto 0); -- local time at datagram rcv time
	T_prime_1 : in std_logic_vector(63 downto 0); -- rcvd RT field
	T_prime_2 : in std_logic_vector(63 downto 0); -- rcvd TT field

	compute_done: out std_logic;
	is_valid: out std_logic;
	theta : out std_logic_vector(63 downto 0);
	delta : out std_logic_vector(63 downto 0);
	epoch  : out std_logic_vector(63 downto 0)
	);
end ntp_theta_delta;

architecture RTL of ntp_theta_delta is
component ntp_addsub is 
generic(
	is_adder : boolean :=true;
	size : natural :=64 -- must be even number here
	);
port(
	clk : in std_logic;

	op_1 : in std_logic_vector(size-1 downto 0);
	op_2 : in std_logic_vector(size-1 downto 0);

	result : out std_logic_vector(size-1 downto 0)
	);
end component ntp_addsub;

constant comp_dly : natural:=3;
signal update_cnt : natural range 0 to comp_dly-1;

signal add_op_1 : std_logic_vector(63 downto 0);
signal add_op_2 : std_logic_vector(63 downto 0);
signal add_result : std_logic_vector(63 downto 0);

signal sub_op_1 : std_logic_vector(63 downto 0);
signal sub_op_2 : std_logic_vector(63 downto 0);
signal sub_result : std_logic_vector(63 downto 0);

-- intermediate results
signal delta_1 : std_logic_vector(63 downto 0);
signal delta_2 : std_logic_vector(63 downto 0);

signal theta_1 : std_logic_vector(63 downto 0);
signal theta_2 : std_logic_vector(63 downto 0);

type def_state_td_comp is (idle, compute_1,compute_2,prepare_theta,
		compute_theta,compute_delta);
signal state_td_comp : def_state_td_comp;

begin
-------------------------------------------------------------------------------
-- computers
-------------------------------------------------------------------------------
lbl_adder : ntp_addsub generic map(
	is_adder=>true,
	size=>64
	)
port map(
	clk=>refclk,

	op_1=>add_op_1,
	op_2=>add_op_2,

	result=>add_result
	);

lbl_subtracter : ntp_addsub generic map(
	is_adder=>false,
	size=>64
	)
port map(
	clk=>refclk,

	op_1=>sub_op_1,
	op_2=>sub_op_2,

	result=>sub_result
	);
-------------------------------------------------------------------------------
-- ntp adder
-------------------------------------------------------------------------------
process(refclk)
begin
	if rising_edge(refclk) then
		if (rst='1') then
			state_td_comp<=idle;
		else
			case state_td_comp is
				when idle =>
					update_cnt<=0;
					if ntp_timeout='1' then
						compute_done<='1';
						theta<=(others=>'0');
						delta<=(others=>'0');
						is_valid<='0';
						epoch<=T_2;

					elsif ntp_frame_rcvd='1' then
						is_valid<='1';
						sub_op_1<=T_2;
						sub_op_2<=T_1;
						
						add_op_1<=T_2;
						add_op_2<=T_1;
						epoch<=T_2;
						state_td_comp<=compute_1;
					else
						compute_done<='0';
					end if;
				
				when compute_1 =>
					if update_cnt=comp_dly-1 then
						--save results
						delta_1<=sub_result;
						theta_2<=SXT(add_result(63 downto 1),64);

						-- initiate new computation
						sub_op_1<=T_prime_2;
						sub_op_2<=T_prime_1;
						
						add_op_1<=T_prime_2;
						add_op_2<=T_prime_1;

						update_cnt<=0;
						state_td_comp<=compute_2;
					else
						update_cnt<=update_cnt+1;
					end if;

				when compute_2 =>
					if update_cnt=comp_dly-1 then
						--save results
						delta_2<=sub_result;
						theta_1<=SXT(add_result(63 downto 1),64);

						update_cnt<=0;
						state_td_comp<=prepare_theta;
					else
						update_cnt<=update_cnt+1;
					end if;

				when prepare_theta =>
					-- initiate new computation
					sub_op_1<=theta_1;
					sub_op_2<=theta_2;
					
					state_td_comp<=compute_theta;

				when compute_theta =>
					if update_cnt=comp_dly-1 then
						--save results
						theta<=sub_result;

						-- initiate new computation
						sub_op_1<=delta_1;
						sub_op_2<=delta_2;

						update_cnt<=0;
						state_td_comp<=compute_delta;
					else
						update_cnt<=update_cnt+1;
					end if;

				when compute_delta =>
					if update_cnt=comp_dly-1 then
						--save results
						delta<=sub_result;

						update_cnt<=0;
						compute_done<='1';
						state_td_comp<=idle;
					else
						update_cnt<=update_cnt+1;
					end if;
			end case;
		end if;
	end if;
end process;
end RTL;
