-- tb_ntp

-- Olivier Bourrion, May 2015

library ieee;
use ieee.std_logic_1164.all;
use IEEE.NUMERIC_STD.ALL;

use work.pack_ipv4_types.all;
use work.pack_axi.all;
use work.pack_arp_types.all;

entity tb_ntp is
generic (
	PC_MAC_NUMBER	: std_logic_vector(47 downto 0):=x"0000_1cb5_ca61";
	PC_IP_ADDRESS   : std_logic_vector(31 downto 0):=X"c0a864" & X"09"; -- 192.168.100.09

	ELEC_MAC_NUMBER : std_logic_vector(47 downto 0):=x"0022_8f03_0001"; --CENTRAL 
	ELEC_IP_ADDRESS   : std_logic_vector(31 downto 0):=X"c0a864" & X"11"; -- 192.168.100.17

-- 	REFCLK_FREQ :  natural:=10E6; -- default to 10MHz
-- 	refclk_period : time :=100 ns; --10 MHz

	REFCLK_FREQ :  natural:=50E6; -- default to 50MHz
	refclk_period : time :=20 ns; -- 50MHz
	poll_interval_exp :  integer:=2; -- 7 for 2^7=128s -- should be between 6 and 10
-- 	clock_precision :  integer:=20 -- defined as 1s/2^-clock precision (20 for ~ 1 us)
	clock_precision :  integer:=22 -- defined as 1s/2^-clock precision (20 for ~ 1 us)
	);
end entity tb_ntp;

architecture simu of tb_ntp is
component ntp_main
generic(
	IN_SIMULATION :  boolean:=false; -- when true, time_enable period is  accelerated by 30000 (3 us instead of 1s)
	REFCLK_FREQ :  natural:=10E6; -- default to 10MHz
	poll_interval_exp :  integer:=7; -- 7 for 2^7=128s -- should be between 6 and 10
	clock_precision :  integer:=20 -- defined as 1s/2^-clock precision (20 for ~ 1 us)
	);
port(
	mac_clk: in std_logic; -- Ethernet MAC clock (125MHz)
	rst_macclk: in std_logic; -- MAC clock domain sync reset

	our_ip_address 	: in std_logic_vector (31 downto 0);
	ntp_server_add 	: in std_logic_vector (31 downto 0);
	enable_ntp : in std_logic;

	-- UDP RX signals
    udp_rx_start : in std_logic;       -- indicates receipt of udp header
    udp_rxo      : in udp_rx_type;

	-- UDP TX signals
	udp_tx_start			: out std_logic;					-- indicates req to tx UDP
	udp_txi					: out udp_tx_type;					-- UDP tx cxns
	udp_tx_result			: in std_logic_vector (1 downto 0); -- tx status (changes during transmission)
	udp_tx_data_out_ready   : in std_logic;						-- indicates udp_tx is ready to take data

	-- UDP TX signals (from next stage)
	udp_tx_start_prior	    : in std_logic;
	udp_txi_prior			: in udp_tx_type;
	udp_tx_result_prior		: out std_logic_vector (1 downto 0);
	udp_tx_data_out_ready_prior: out std_logic;

	-- refclk interface (xtime slower than mac_clk)
	refclk: in std_logic;
	rst_refclk: in std_logic;

	-- time and monitoring (refclk sync)
	new_time_avail: out std_logic;-- one refclk pulse on each update
	local_time : out std_logic_vector(63 downto 0);

	new_theta_delta: out std_logic; -- one refclk pulse on each update
	theta_i : out std_logic_vector(63 downto 0);
	delta_i : out std_logic_vector(63 downto 0);

	update_theta: out std_logic; -- one refclk pulse on each update
	theta : out std_logic_vector(63 downto 0);
	delta : out std_logic_vector(63 downto 0);
	freq_adj : out std_logic_vector(clock_precision-1 downto 0)
	);
end component;

component ntp_udp_server_sim
generic(
	poll_interval_exp :  integer:=7; -- 7 for 2^7=128s -- should be between 6 and 10
	clock_precision :  integer:=20 -- defined as 1s/2^-clock precision (20 for ~ 1 us)
	);
port(
	mac_clk: in std_logic; -- Ethernet MAC clock (125MHz)
	rst_macclk: in std_logic; -- MAC clock domain sync reset

	our_ip_address 	: in std_logic_vector (31 downto 0);
	ntp_client_add 	: in std_logic_vector (31 downto 0);

	-- UDP RX signals
    udp_rx_start : in std_logic;       -- indicates receipt of udp header
    udp_rxo      : in udp_rx_type;

	-- UDP TX signals
	udp_tx_start			: out std_logic;					-- indicates req to tx UDP
	udp_txi					: out udp_tx_type;					-- UDP tx cxns
	udp_tx_result			: in std_logic_vector (1 downto 0); -- tx status (changes during transmission)
	udp_tx_data_out_ready   : in std_logic;						-- indicates udp_tx is ready to take data

	-- UDP TX signals (from next stage)
	udp_tx_start_prior	    : in std_logic;
	udp_txi_prior			: in udp_tx_type;
	udp_tx_result_prior		: out std_logic_vector (1 downto 0);
	udp_tx_data_out_ready_prior: out std_logic;

	-- refclk interface (xtime slower than mac_clk)
	refclk: in std_logic
	);
end component;

component UDP_Complete
generic (
	IN_SIMULATION     : boolean :=false;
	SIMUL_DST_MAC_NUMBER  : std_logic_vector(47 downto 0):=x"AAAA_BBBB_CCCC";
	CLOCK_FREQ			: integer := 125000000;							-- freq of data_in_clk -- needed to timout cntr
	ARP_TIMEOUT			: integer := 60;									-- ARP response timeout (s)
	ARP_MAX_PKT_TMO	: integer := 5;									-- # wrong nwk pkts received before set error
	MAX_ARP_ENTRIES 	: integer := 255									-- max entries in the ARP store
	);
Port (
	-- UDP TX signals
	udp_tx_start			: in std_logic;							-- indicates req to tx UDP
	udp_txi					: in udp_tx_type;							-- UDP tx cxns
	udp_tx_result			: out std_logic_vector (1 downto 0);-- tx status (changes during transmission)
	udp_tx_data_out_ready: out std_logic;							-- indicates udp_tx is ready to take data

	-- UDP RX signals
	udp_rx_start			: out std_logic;							-- indicates receipt of udp header
	udp_rxo					: out udp_rx_type;

	-- IP RX signals
	ip_rx_hdr				: out ipv4_rx_header_type;

	-- system signals
	clk125					: in  STD_LOGIC;
	reset 					: in  STD_LOGIC;
	our_ip_address 		: in STD_LOGIC_VECTOR (31 downto 0);
	our_mac_address 		: in std_logic_vector (47 downto 0);
	control					: in udp_control_type;

	-- status signals
	arp_pkt_count			: out STD_LOGIC_VECTOR(7 downto 0);			-- count of arp pkts received
	ip_pkt_count			: out STD_LOGIC_VECTOR(7 downto 0);			-- number of IP pkts received for us
	
	-- gmii TX interface
	-- 	gmii_gtx_clk : out STD_LOGIC; -- must be generated with appropriate system
	gmii_tx_en : out STD_LOGIC;
	gmii_tx_er : out STD_LOGIC;
	gmii_txd : out STD_LOGIC_VECTOR(7 downto 0);

	-- gmii RX interface
	gmii_rx_clk: in STD_LOGIC;
   	gmii_rx_dv: in STD_LOGIC;
	gmii_rx_er: in STD_LOGIC;
	gmii_rxd : in STD_LOGIC_VECTOR(7 downto 0)
	);
end component;
signal mac_clk: std_logic; -- Ethernet MAC clock (125MHz)
signal rst_macclk: std_logic; -- MAC clock domain sync reset
signal refclk: std_logic;
signal rst_refclk: std_logic; -- referenced to refclk

signal gmii_tx_en_from_client :std_logic;
signal gmii_tx_er_from_client :std_logic;
signal gmii_txd_from_client :std_logic_vector(7 downto 0);

signal gmii_tx_en_from_server :std_logic;
signal gmii_tx_er_from_server :std_logic;
signal gmii_txd_from_server :std_logic_vector(7 downto 0);

signal udp_rx_start : std_logic;       -- indicates receipt of udp header
signal udp_rxo      : udp_rx_type;
signal udp_tx_start			: std_logic;					-- indicates req to tx UDP
signal udp_txi					: udp_tx_type;					-- UDP tx cxns
signal udp_tx_result			: std_logic_vector (1 downto 0); -- tx status (changes during transmission)
signal udp_tx_data_out_ready   : std_logic;						-- indicates udp_tx is ready to take data
signal control			: udp_control_type;

signal udp_tx_start_prior	    : std_logic;
signal udp_txi_prior			: udp_tx_type;

-- server side
signal udp_rx_start_server : std_logic;       -- indicates receipt of udp header
signal udp_rxo_server      : udp_rx_type;
signal udp_tx_start_server			: std_logic;					-- indicates req to tx UDP
signal udp_txi_server					: udp_tx_type;					-- UDP tx cxns
signal udp_tx_result_server			: std_logic_vector (1 downto 0); -- tx status (changes during transmission)
signal udp_tx_data_out_ready_server   : std_logic;						-- indicates udp_tx is ready to take data


begin
-------------------------------------------------------------------------------
-- clocking
-------------------------------------------------------------------------------
process
begin
	mac_clk<='0';
	wait for 4 ns;
	
	mac_clk<='1';
	wait for 4 ns;
end process;

process
begin
	refclk<='0';
	wait for refclk_period/2;
	
	refclk<='1';
	wait for refclk_period/2;
end process;
-------------------------------------------------------------------------------
-- Client side
-------------------------------------------------------------------------------
lbl_ntp_main : ntp_main generic map(
	IN_SIMULATION=>true,
	REFCLK_FREQ=>REFCLK_FREQ,
	poll_interval_exp=>poll_interval_exp,
	clock_precision=>clock_precision
	)
port map(
	mac_clk=>mac_clk,
	rst_macclk=>rst_macclk,

	our_ip_address=>ELEC_IP_ADDRESS,
	ntp_server_add=>PC_IP_ADDRESS,
	enable_ntp=>'1',

	-- UDP RX signals
    udp_rx_start=>udp_rx_start,
    udp_rxo=>udp_rxo,

	-- UDP TX signals
	udp_tx_start=>udp_tx_start,
	udp_txi=>udp_txi,
	udp_tx_result=>udp_tx_result,
	udp_tx_data_out_ready=>udp_tx_data_out_ready,

	-- UDP TX signals (from next stage)
	udp_tx_start_prior=>udp_tx_start_prior,
	udp_txi_prior=>udp_txi_prior,
	udp_tx_result_prior=>open,
	udp_tx_data_out_ready_prior=>open,

	-- refclk interface (xtime slower than mac_clk)
	refclk=>refclk,
	rst_refclk=>rst_refclk,

	-- time and monitoring (refclk sync)
	new_time_avail=>open,
	local_time=>open,
	
	new_theta_delta=>open,
	theta_i=>open,
	delta_i=>open,


	update_theta=>open,
	theta=>open,
	delta=>open,
	freq_adj=>open
	);
udp_tx_start_prior<='0';

lbl_UDP_client : UDP_Complete generic map(
		IN_SIMULATION=>true,
		SIMUL_DST_MAC_NUMBER=>ELEC_MAC_NUMBER,
		CLOCK_FREQ=>125000000,
		ARP_TIMEOUT=>60,
		ARP_MAX_PKT_TMO=>5,
		MAX_ARP_ENTRIES=>255
		)
	port map(
		-- UDP TX signals
		udp_tx_start=>udp_tx_start,
		udp_txi=>udp_txi,
		udp_tx_result=>udp_tx_result,
		udp_tx_data_out_ready=>udp_tx_data_out_ready,
		
		-- UDP RX signals
		udp_rx_start=>udp_rx_start,
		udp_rxo=>udp_rxo,
		
		-- IP RX signals
		ip_rx_hdr=>open,

		-- system signals
		clk125=>mac_clk,
		reset=>rst_macclk,
		our_ip_address=>ELEC_IP_ADDRESS,
		our_mac_address=>ELEC_MAC_NUMBER,
		control=>control,

		-- status signals
		arp_pkt_count=>open,
		ip_pkt_count=>open,
		
		-- gmii TX interface
		-- 	gmii_gtx_clk : out STD_LOGIC; -- must be generated with appropriate system
		gmii_txd=>gmii_txd_from_client,
		gmii_tx_en=>gmii_tx_en_from_client,
		gmii_tx_er=>gmii_tx_er_from_client,

		-- gmii RX interface
		gmii_rx_clk=>mac_clk,
		gmii_rxd =>gmii_txd_from_server,
		gmii_rx_dv=>gmii_tx_en_from_server,
		gmii_rx_er=>gmii_tx_er_from_server
		);

control.ip_controls.arp_controls.clear_cache<='0';
-------------------------------------------------------------------------------
-- Simulated server side
-------------------------------------------------------------------------------
lbl_ntp_server_sim : ntp_udp_server_sim generic map(
	poll_interval_exp=>poll_interval_exp,
	clock_precision=>clock_precision
	)
port map(
	mac_clk=>mac_clk,
	rst_macclk=>rst_macclk,

	our_ip_address=>PC_IP_ADDRESS,
	ntp_client_add=>ELEC_IP_ADDRESS,

	-- UDP RX signals
    udp_rx_start=>udp_rx_start_server,
    udp_rxo=>udp_rxo_server,

	-- UDP TX signals
	udp_tx_start=>udp_tx_start_server,
	udp_txi=>udp_txi_server,
	udp_tx_result=>udp_tx_result_server,
	udp_tx_data_out_ready=>udp_tx_data_out_ready_server,

	-- UDP TX signals (from next stage)
	udp_tx_start_prior=>udp_tx_start_prior,
	udp_txi_prior=>udp_txi_prior,
	udp_tx_result_prior=>open,
	udp_tx_data_out_ready_prior=>open,

	-- refclk interface (xtime slower than mac_clk)
	refclk=>refclk
	);

lbl_UDP_server : UDP_Complete generic map(
		IN_SIMULATION=>true,
		SIMUL_DST_MAC_NUMBER=>ELEC_MAC_NUMBER,
		CLOCK_FREQ=>125000000,
		ARP_TIMEOUT=>60,
		ARP_MAX_PKT_TMO=>5,
		MAX_ARP_ENTRIES=>255
		)
	port map(
		-- UDP TX signals
		udp_tx_start=>udp_tx_start_server,
		udp_txi=>udp_txi_server,
		udp_tx_result=>udp_tx_result_server,
		udp_tx_data_out_ready=>udp_tx_data_out_ready_server,
		
		-- UDP RX signals
		udp_rx_start=>udp_rx_start_server,
		udp_rxo=>udp_rxo_server,
		
		-- IP RX signals
		ip_rx_hdr=>open,

		-- system signals
		clk125=>mac_clk,
		reset=>rst_macclk,
		our_ip_address=>PC_IP_ADDRESS,
		our_mac_address=>PC_MAC_NUMBER,
		control=>control,

		-- status signals
		arp_pkt_count=>open,
		ip_pkt_count=>open,
		
		-- gmii TX interface
		-- 	gmii_gtx_clk : out STD_LOGIC; -- must be generated with appropriate system
		gmii_txd=>gmii_txd_from_server,
		gmii_tx_en=>gmii_tx_en_from_server,
		gmii_tx_er=>gmii_tx_er_from_server,

		-- gmii RX interface
		gmii_rx_clk=>mac_clk,
		gmii_rxd =>gmii_txd_from_client,
		gmii_rx_dv=>gmii_tx_en_from_client,
		gmii_rx_er=>gmii_tx_er_from_client
		);
-------------------------------------------------------------------------------
-- main
-------------------------------------------------------------------------------
main : process
begin
	rst_refclk<='1';
	rst_macclk<='1';
	wait for 300 ns;

	rst_refclk<='0';
	rst_macclk<='0';
	wait for 10 ns;

	wait for 600 us;
	report "fin de simulation" severity error;

end process main;

end simu;
